/**************************************************************
 * SACHIP: A verilog IP Core to implement the Sequence Alignment
 * Algorithm in Hardware. This design is mostly structural and
 * hence can be synthesized using any synthesis tools like 
 * DesignCompiler or Encounter RTL Compiler.
 *
 * Please report problems to vamsi.krishnak@gmail.com
 **************************************************************/


/* StringRegister: String Register stores s
 * it returns s[index] into out.
 * */
module StringRegister8X(s,index,out);
input [63:0]s;
/*8-bit unsigned integer*/
input [7:0]index;
output [7:0]out;
wire [63:0]s;
reg [7:0]out;

always @(index or s) begin
 case(index)
 	 0: out ="";
 	 1: out = s[7:0];
  	 2: out = s[15:8];
  	 3: out = s[23:16];
   	 4: out = s[31:24];
   	 5: out = s[39:32];
   	 6: out = s[47:40];
   	 7: out = s[55:48];
   	 8: out = s[63:56];
  endcase
end
endmodule


module CharXORBlock(c1,c2,out);
input [7:0]c1; 
output [7:0]out;
input [7:0]c2;
wire [7:0]c1; wire [7:0]c2;
reg [7:0]out;
always @(c1 or c2) begin
	if(c1 == c2) 
	begin
		out = 0;
	end
	else 
	begin
		out = 1;
	end
end
endmodule

/*Adds j to i*/
module AddBlock(i,j,out);
input [7:0]i;wire [7:0]i;
input [7:0]j;wire [7:0]j;
output [7:0]out; wire [7:0]out;
assign	out = i+j;
endmodule

module Add1Block(i,out);
input [7:0]i;
output [7:0]out;
wire [7:0]out;
wire [7:0]i;
assign out = i+1;
endmodule

module Comparator2Block(i,j,out);
input [7:0]i; wire [7:0]i;
input [7:0]j; wire [7:0]j;
output [7:0]out; reg [7:0]out;
always @(i or j) begin
	if(i < j) begin
		out = i;
    end
	else begin
		out = j;
	end
end
endmodule

module Comparator3Block(d_i1_j1,d_i1_j,d_i_j1,out);
input [7:0]d_i1_j1;
input [7:0]d_i1_j;
input [7:0]d_i_j1; output [7:0]out;
wire  [7:0]net1;
Comparator2Block CMP1 (d_i1_j1,d_i1_j,net1);
Comparator2Block CMP2 (net1,d_i_j1,out);
endmodule
/*
 * ComputeBlock: Computes the core minmum of
 * the dynamic programming at every stage.
 * D(i,j) = min {D(i-1,j-1)+c(c1,c2), D(i-1,j)+1,D(i,j-1)+1}
 * */
module ComputeBlock(c1,c2,d_i1_j1,d_i1_j,d_i_j1,d_i_j);
input [7:0]c1;wire [7:0]c1;
input [7:0]c2;wire [7:0]c2;
input [7:0]d_i1_j1;wire [7:0]d_i1_j1;
input [7:0]d_i_j1;wire [7:0]d_i_j1;
input [7:0]d_i1_j;wire [7:0]d_i1_j;
output [7:0]d_i_j; wire [7:0]net1;
wire [7:0]net2;wire [7:0]net3; wire [7:0]net4;

CharXORBlock cmp_xor (c1,c2,net1); 
AddBlock cmp_add (d_i1_j1,net1,net2);
Add1Block cmp_add1 (d_i1_j,net3);
Add1Block cmp_add2 (d_i_j1,net4);
Comparator3Block cmp3d (net2,net3,net4,d_i_j);
endmodule

module SimpleMUX1(in,s1,s2,out);
input in;input [7:0]s1;input [7:0]s2;
output [7:0]out; reg [7:0]out;

always @(in or s1 or s2) begin
	if(in) begin
		out = s1;
	end
	else
		out = s2;
	begin
	end
end
endmodule

/*
 * AlgoShifter: Customized shifter to implement 
 * the sequence alignment algorithm. This is the 
 * core of the algorithm, initially we write 
 * the previous row as follows, as we compute D(i,j)
 * we shift that into this shift register at the end
 * as shown below, so after n*8 shifts D(i,j) will
 * come to the begining. 
 *
 * [D(i-1,j-1),D(i-1,j)...prev_row...][...current_row......D(i,j)]
 *
 * Finally if we shift n*n*8 bits O(8*n^2) bits the final 
 * edit distance is computed.
 *
 * AlgoShifter(reset,stop,d_i_j,d_i_0,out1,out2)
 * out1 = D(i-1,j-1)
 * out2 = D(i-1,j)
 * 
 * when reset is used 
 */
module AlgoShifter8X (reset,stop,d_i_j,d_i_0,out1,out2,out3,
	clk,restart);
input reset;input [7:0]d_i_j;input stop;input [7:0]d_i_0;
input restart;
output [7:0]out1;
output [7:0]out2;
output [7:0]out3;
input clk;
reg [79:0]store;
wire [7:0]out1;
wire [7:0]out2;
wire [7:0]out3;
wire reset;wire [7:0]d_i_j;wire [7:0]d_i_0;
wire stop; wire [7:0]mux_out;
wire clk;

always @(posedge restart) begin
 store[7:0]   = 0;
 store[15:8]  = 1;
 store[23:16] = 2;
 store[31:24] = 3;
 store[39:32] = 4;
 store[47:40] = 5;
 store[55:48] = 6;
 store[63:56] = 7;
 store[71:64] = 8;
 store[79:72] = 1;
end

always @ (posedge clk) begin
	if(stop==0) begin
		    /* << operator seems not working here*/
 			store[7:0]   = store[15:8];
 			store[15:8]  = store[23:16];
 			store[23:16] = store[31:24];
 			store[31:24] = store[39:32];
 			store[39:32] = store[47:40];
 			store[47:40] = store[55:48];
 			store[55:48] = store[63:56];
 			store[63:56] = store[71:64];
			store[71:64] = store[79:72];
			if(reset) begin
				store[79:72] = d_i_0;
			end
			else begin
				store[79:72] = d_i_j;
			end
	end
end
  assign out1 = store[7:0];
  assign out2 = store[15:8];
  assign out3 = store[79:72];
endmodule

/*Combinatorial circuit for the STOP signal*/
module StopSignalBlock(i,j,stop);
input [7:0]i;input [7:0]j;output stop;
reg stop;
always @(i or j) begin 
	if(i==8 && j==8) 
	begin
		stop = 1;
	end
	else 
	begin
		stop = 0;
	end
end
endmodule

/****
 * CounterBlock8X: Generates the control
 * signals also generates a restart and 
 * stop signals
 */
module CounterBlock8X(i,j,reset,reset_input,clk,restart,stop);
input clk; input restart; output stop;
output [7:0]i;
output [7:0]j; output reset; output [7:0]reset_input;
wire [7:0]reset_input;
reg [7:0]i; wire stop;
reg [7:0]j;
reg reset;
always @ (posedge clk or posedge restart) begin
	if(restart==1) 
	begin
		i = 1;
		j = 1;
		reset=0;
	end
	else  
	begin
		if(i == 8) 
		begin
			if(j != 8) 
			begin
				reset=1;
				i = 0;
				j = j +1;
			end
        end
		else 
		begin
			reset=0;
			i = i+1;
		end
	end
end
assign reset_input = j;
StopSignalBlock stp1(i,j,stop);
endmodule
/*
 * AlignmentAlgo: Computes the 
 * cost of minimum EditDistance between
 * string s1 and s2 the value minimum EditiDistance
 * is given as output.
 * 8X currently it aligns strings of length 8(exactly).
 * out is a 8-bit integer holding the value.
 * */
module AlignmentAlgo8X(s1,s2,d_i_j,clk,restart);
input clk;input [63:0]s2;
input [63:0]s1;wire [63:0]s2;
input restart;
output [7:0]d_i_j;wire [7:0]d_i_j;
wire [63:0]s1;
wire clk; wire gated_clk;
wire [7:0]c1;wire [7:0]c2;wire [7:0]d_i1_j1;
wire [7:0]d_i1_j;wire [7:0]d_i_j1;wire [7:0]i;wire [7:0]j;
wire reset; wire [7:0]reset_input;wire stop;
CounterBlock8X control_block (i,j,reset,reset_input,gated_clk,
	restart,stop);
StringRegister8X sr1 (s1,i,c1);
StringRegister8X sr2 (s2,j,c2);
ComputeBlock cmp1 (c1,c2,d_i1_j1,d_i1_j,d_i_j1,d_i_j);
AlgoShifter8X shft1 (reset,stop,d_i_j,reset_input,d_i1_j1,
	d_i1_j,d_i_j1,gated_clk,restart);
/*Gate the clock 
 * 1. once the stop signal is set...
 * 2. also stop every thing during restart...
 **/
assign gated_clk = (clk&(~stop))&(~restart);
endmodule 
